import re
from randomlib.tokenizer import Tokenize
# from importlib.resources import files
# import randomlib.preprocess
from importlib_resources import files

class Preprocess:
    stopwords = []

    def __init__(self):
        if not Preprocess.stopwords:
            Lines = files('randomlib.preprocess').joinpath(
                'marathiStopwords.txt').read_text(encoding="utf8").split('\n')
            for line in Lines:
                Preprocess.stopwords.append(line.strip())

    def removeUrl(self, text):
        return re.sub(r"http\S+", "", text)

    def removeStopwords(self, text):
        newlist = []
        t = Tokenize()
        textlist = t.wordTokenize(text, False)
        for word in textlist:
            if word not in Preprocess.stopwords:
                newlist.append(word)
        return newlist

    def removeNondevnagari(self, line, engNum = 1):
        line = [i for i in line]
        chars = ''' क ख ग घ ङ च छ ज झ ञ ट ठ ड ढ ण त थ द ध न प फ ब भ म य र ल व र्‍ श ष स ह क़ ख़ ग़ ऩ ड़ ढ़ ऱ य़ ळ ऴ फ़ ज़ ॹ ॺ ॻ ॼ ॾ ॿ ् ऄ अ आ इ ई उ ऊ ॶ ॷ ऋ ॠ ऌ ॡ ॲ ॕ ा ि ी ु ू ॖ ॗ ृ ॄ ॢ ॣ ऍ ऎ ए ऐ ऑ ऒ ओ औ ॵ ॳ ॴ ॅ ॆ े ै ॉ ॊ ो ौ ॏ ऺ ऻ ॎ ॐ ँ ऀ ं ॱ ः ॑ ॒ ॓ ॔ ऽ ॽ , . ० १ २ ३ ४ ५ ६ ७ ८ ९ ₹ । ॥ | ॰ '''
        engNums = '''0123456789'''
        i = 0
        while i < len(line):
            if engNum == 1:
                if ((line[i] not in chars) and (line[i] not in engNums)):
                    del line[i]
                    i -= 1
                i += 1
            elif engNum == 0:
                if (line[i] not in chars):
                    del line[i]
                    i -= 1
                i += 1
            else:
                print(
                    "Please pass the correct argument. \n 1 for retaining English numbers \n 0 for not retaining English numbers")
                break
        if (line[0] == " "):
            del line[0]
        if (line[len(line)-1] == " "):
            del line[len(line)-1]
        print("".join(line))



from randomlib.modelRepo.mahaSimilarity import SimilarityModel

class SimilarityAnalyzer(SimilarityModel):
    def __init__(self, modelName = 'marathi-sentence-similarity-sbert'):
        self.modelName = modelName
        super().__init__(self.modelName)

    def embedSentences(self, sentences):
        return super().embedSentences(sentences)
    
    def getSimilarityScore(self, sourceSentence, sentences, as_dict: bool = False):
        return super().getSimilarityScore(sourceSentence, sentences, as_dict)

    def listModels(self):
        return super().listModels()


from .mahaFill import MaskFillModel
from .mahaGPT import GPTModel
from .mahaHate import HateModel
from .mahaNER import NERModel
from .mahaSentiment import SentimentModel
from .mahaSimilarity import SimilarityModel


import logging
import huggingface_hub.utils as hf_hub_utils
import pandas as pd
logging.disable(logging.INFO) # disable INFO 
logging.disable(logging.DEBUG) # disable INFO 
logging.disable(logging.WARNING) # disable INFO 
hf_hub_utils.disable_progress_bars()
pd.options.display.max_colwidth = None




from randomlib.modelRepo.mahaFill import MaskFillModel

class MaskPredictor(MaskFillModel):
    def __init__(self, modelName = 'marathi-bert-v2'):
        self.modelName = modelName
        super().__init__(self.modelName)

    def predictMask(self, text, details: str = "minimum", as_dict: bool = False):
        return super().predictMask(text, details, as_dict)

    def listModels(self):
        return super().listModels()


paths = { 
    "autocomplete": {"marathi-gpt": "l3cube-pune/marathi-gpt"},

    "hate": {"mahahate-bert": "l3cube-pune/mahahate-bert",
             "mahahate-multi-roberta": "l3cube-pune/mahahate-multi-roberta"},
    
    "maskFill": {"marathi-bert-v2": "l3cube-pune/marathi-bert-v2",
                 "marathi-roberta": "l3cube-pune/marathi-roberta",
                 "marathi-albert": "l3cube-pune/marathi-albert"},
    
    "sentiment": {"MarathiSentiment": "l3cube-pune/MarathiSentiment"},

    "similarity": {"marathi-sentence-similarity-sbert": "l3cube-pune/marathi-sentence-similarity-sbert",
                   "marathi-sentence-bert-nli": "l3cube-pune/marathi-sentence-bert-nli"},
   
    "tagger": {"marathi-ner": "l3cube-pune/marathi-ner"}
    
}



from randomlib.modelRepo.mahaHate import HateModel

class HateAnalyzer(HateModel):
    def __init__(self, modelName = 'mahahate-bert'):
        self.modelName = modelName
        super().__init__(self.modelName)

    def getPolarityScore(self, text):
        return super().getPolarityScore(text)

    def listModels(self):
        return super().listModels()



from transformers import AutoTokenizer, AutoModelForSequenceClassification
from transformers import pipeline
import pandas as pd
from ..config import paths

class SentimentModel:
    def __init__(self, modelName='MarathiSentiment'):
        self.modelName = modelName
        self.modelRoute = paths['sentiment'][self.modelName]
        self.tokenizer = AutoTokenizer.from_pretrained(self.modelRoute)
        self.model = AutoModelForSequenceClassification.from_pretrained(self.modelRoute)
        self.classifier = pipeline('text-classification',
                              model=self.model, tokenizer=self.tokenizer)
    def getPolarityScore(self, text):
        result = self.classifier(text)
        df = pd.DataFrame.from_dict(result)
        return df

    def listModels(self):
        print(" sentiment models: ")
        for model in paths['sentiment']:
            print("\t",model, ": ", paths['sentiment'][model])
        for task in set(paths) - {'sentiment'}:
            print("\n",task,"models: ")
            for model in paths[task]:
                print("\t",model, ": ", paths[task][model])


from randomlib.modelRepo.mahaGPT import GPTModel

class TextGenerator(GPTModel):
    def __init__(self, modelName = 'marathi-gpt'):
        self.modelName = modelName
        super().__init__(self.modelName)

    def nextWord(self, text, numOfPredictions = 1):
        return super().nextWord(text, numOfPredictions)

    def completeSentence(self, text, numOfWords = 25, numOfPredictions = 1):
        return super().completeSentence(text, numOfWords, numOfPredictions )

    def listModels(self):
        return super().listModels()


from randomlib.modelRepo.mahaNER import NERModel

class EntityRecognizer(NERModel):

    def __init__(self,modelName:str = 'marathi-ner',
                gpu_enabled: bool=False) -> None:
        self.modelName = modelName
        super().__init__(self.modelName,gpu_enabled)

    def getPolarityScore(self, text, details: str = "minimum", as_dict: bool = False):
        return super().getPolarityScore(text, details, as_dict)

    def getTokenLabels(self, text):
        return super().getTokenLabels(text)

    def listModels(self):
        return super().listModels()


from sentence_transformers import SentenceTransformer, util
import numpy
from ..config import paths

class SimilarityModel:
    def __init__(self, modelName = 'marathi-sentence-similarity-sbert'):
        self.modelName = modelName
        self.modelRoute = paths['similarity'][self.modelName]
        self.model = SentenceTransformer(self.modelRoute)

    def embedSentences(self,sentences):
        sentenceEmbeddings = self.model.encode(sentences)
        return sentenceEmbeddings

    def getSimilarityScore(self, sourceSentence, sentences, as_dict: bool = False):
        embeddings1 = self.embedSentences(sourceSentence)
        embeddings2 = self.embedSentences(sentences)
        cosine_scores = util.cos_sim(embeddings1, embeddings2)
        resultNpArray = cosine_scores.numpy()[0]

        if as_dict:
          dict = {}
          if type(sentences) == str:
            sentences = [sentences]
          for i in range(len(sentences)):
            dict[sentences[i]] =  resultNpArray[i]
          return dict
        return resultNpArray

    def listModels(self):
        print(" similarity models: ")
        for model in paths['similarity']:
            print("\t",model, ": ", paths['similarity'][model])
        for task in set(paths) - {'similarity'}:
            print("\n",task,"models: ")
            for model in paths[task]:
                print("\t",model, ": ", paths[task][model])


    



import pandas as pd
import os
# IMPORTANT! pip install openpyxl

##INTERNAL Function, not meant for programmer's usage
def checkdir(modelname):
    rootPath = os.path.expanduser('~\.cache/')
    if not os.path.isdir(rootPath):
        os.makedirs(os.getcwd()+'/.cache/randomlib/'+modelname,exist_ok=True)
        return os.getcwd()
    else:
        rootPath = os.path.expanduser('~\.cache/')
        os.chdir(rootPath)
        os.makedirs('randomlib/'+modelname,exist_ok=True)
        return '~'

##INTERNAL Function, not meant for programmer's usage
def downloadMahasent():
    dataset_type = {
        'tweets-train.csv': 'https://raw.githubusercontent.com/l3cube-pune/MarathiNLP/main/L3CubeMahaSent%20Dataset/tweets-train.csv',
        'tweets-test.csv': 'https://raw.githubusercontent.com/l3cube-pune/MarathiNLP/main/L3CubeMahaSent%20Dataset/tweets-test.csv',
        'tweets-valid.csv': 'https://raw.githubusercontent.com/l3cube-pune/MarathiNLP/main/L3CubeMahaSent%20Dataset/tweets-valid.csv',
        'tweets-extra.csv': 'https://raw.githubusercontent.com/l3cube-pune/MarathiNLP/main/L3CubeMahaSent%20Dataset/tweets-extra.csv'
    }
    result = {}
    rootPath = os.path.expanduser(checkdir('mahaSent'))
    childPath = os.path.expanduser(rootPath+'/.cache/randomlib/mahaSent')
    os.chdir(childPath)
    for key in dataset_type:
        df = pd.read_csv(dataset_type[key])
        csv_df = df.to_csv(key, index=False, encoding='UTF-16')
        result[key.split(".")[0]] = df  # add df to an dictionary
    os.chdir(rootPath)
    return result

##INTERNAL Function, not meant for programmer's usage
def downloadMahaner():
    iob = {
        'train_iob.txt': 'https://github.com/l3cube-pune/MarathiNLP/blob/main/L3Cube-MahaNER/IOB/train_iob.txt?raw=true',
        'test_iob.txt': 'https://raw.githubusercontent.com/l3cube-pune/MarathiNLP/main/L3Cube-MahaNER/IOB/test_iob.txt',
        'valid_iob.txt': 'https://raw.githubusercontent.com/l3cube-pune/MarathiNLP/main/L3Cube-MahaNER/IOB/valid_iob.txt',
    }
    non_iob = {
        'train_noniob.txt': 'https://github.com/l3cube-pune/MarathiNLP/blob/main/L3Cube-MahaNER/NON_IOB/train_noniob.txt?raw=true',
        'test_noniob.txt': 'https://raw.githubusercontent.com/l3cube-pune/MarathiNLP/main/L3Cube-MahaNER/NON_IOB/test_noniob.txt',
        'valid_noniob.txt': 'https://raw.githubusercontent.com/l3cube-pune/MarathiNLP/main/L3Cube-MahaNER/NON_IOB/valid_noniob.txt',
    }
    result = {'iob': {}, 'non_iob': {}}

    rootPath = os.path.expanduser(checkdir('mahaNER/IOB'))
    rootPath = os.path.expanduser(checkdir('mahaNER/NON_IOB'))

    childPath = os.path.expanduser(rootPath+'/.cache/randomlib/mahaNER/IOB')
    os.chdir(childPath)
    for key in iob:
        df = pd.read_csv(iob[key], sep=" ")
        txt_df = df.to_csv(key, index=False, encoding='UTF-16')
        result['iob'][key.split(".")[0]] = df  # add df to an dictionary

    childPath = os.path.expanduser(rootPath+'/.cache/randomlib/mahaNER/NON_IOB')
    os.chdir(childPath)
    for key in non_iob:
        df = pd.read_csv(non_iob[key], sep="\t")
        txt_df = df.to_csv(key, index=False, encoding='UTF-16', sep="\t")
        result['non_iob'][key.split(".")[0]] = df  # add df to an dictionary
    
    os.chdir(rootPath)
    return result

##INTERNAL Function, not meant for programmer's usage
def downloadMahahate():
    class_2 = {
        'hate_bin_train.xlsx': 'https://github.com/l3cube-pune/MarathiNLP/blob/main/L3Cube-MahaHate/2-class/hate_bin_train.xlsx?raw=true',
        'hate_bin_test.xlsx': 'https://github.com/l3cube-pune/MarathiNLP/blob/main/L3Cube-MahaHate/2-class/hate_bin_test.xlsx?raw=true',
        'hate_bin_valid.xlsx': 'https://github.com/l3cube-pune/MarathiNLP/blob/main/L3Cube-MahaHate/2-class/hate_bin_valid.xlsx?raw=true',
    }
    class_4 = {
        'hate_train.xlsx': 'https://github.com/l3cube-pune/MarathiNLP/blob/main/L3Cube-MahaHate/4-class/hate_train.xlsx?raw=true',
        'hate_test.xlsx': 'https://github.com/l3cube-pune/MarathiNLP/blob/main/L3Cube-MahaHate/4-class/hate_test.xlsx?raw=true',
        'hate_valid.xlsx': 'https://github.com/l3cube-pune/MarathiNLP/blob/main/L3Cube-MahaHate/4-class/hate_valid.xlsx?raw=true',
    }
    result = {'2-class': {}, '4-class': {}}

    rootPath = os.path.expanduser(checkdir('mahaHate/2-class'))
    rootPath = os.path.expanduser(checkdir('mahaHate/4-class'))

    childPath = os.path.expanduser(rootPath+'/.cache/randomlib/mahaHate/2-class')
    os.chdir(childPath)
    for key in class_2:
        df = pd.read_excel(class_2[key])
        xlsx_df = df.to_excel(key, index=False, encoding='UTF-16')
        result['2-class'][key.split(".")[0]] = df  # add df to an dictionary

    childPath = os.path.expanduser(rootPath+'/.cache/randomlib/mahaHate/4-class')
    os.chdir(childPath)
    for key in class_4:
        df = pd.read_excel(class_4[key])
        xlsx_df = df.to_excel(key, index=False, encoding='UTF-16')
        result['4-class'][key.split(".")[0]] = df  # add df to an dictionary
    os.chdir(rootPath)
    return result


def loadDatasets(name):

    if name == 'mahaSent':
        res_dict = downloadMahasent()
        return res_dict
    elif name == 'mahaHate':
        res_dict = downloadMahahate()
        return res_dict
    elif name == 'mahaNER':
        res_dict = downloadMahaner()
        return res_dict



from transformers import BertTokenizerFast, BertForTokenClassification, TokenClassificationPipeline
from IPython.display import display
import torch
import logging
import pandas as pd
from ..config import paths

################# LOGGING #################
nerPipeLogger = logging.getLogger(__name__)
consoleHandler = logging.StreamHandler()
logFormatter = logging.Formatter(fmt=' %(name)s :: %(levelname)-4s :: %(message)s')

nerPipeLogger.setLevel(logging.DEBUG)
consoleHandler.setLevel(logging.DEBUG)

consoleHandler.setFormatter(logFormatter)
nerPipeLogger.addHandler(consoleHandler)

################# LOGGING #################

class NERModel:
    def __init__(self, modelName='marathi-ner',gpu_enabled:bool = False):
        self.modelName = modelName
        self.modelRoute = paths['tagger'][self.modelName]

        self.gpu_enabled = gpu_enabled
        self.device = 1 if (self.gpu_enabled and torch.cuda.is_available()) else -1

        try:
            self.pretrainedNERModel = BertForTokenClassification.from_pretrained(self.modelRoute)
        except Exception as e:
            nerPipeLogger.exception(
                msg="some error has occured while Loading mahaNER_BERT Model", exc_info=e)
            return None

        try:
            self.NERTokenizer = BertTokenizerFast.from_pretrained(self.modelRoute)
        except Exception as e:
            nerPipeLogger.exception(
                msg="some error has occured while Loading mahaNER_BERT Tokenizer", exc_info=e)
            return None

        self.pipeline = TokenClassificationPipeline(
            task='marathi-ner',
            model=self.pretrainedNERModel,
            tokenizer=self.NERTokenizer,
            framework="pt",
            aggregation_strategy='first',
            device=self.device,
        )

    def getPolarityScore(self, text, details: str = "minimum",as_dict:bool = False):
        '''
        text = A string of raw text as a input to NER model
        details = Defines the level of details to get from the prediction.
                possible values = 'minimum' (default),'medium','all'.
        as_dict = returns the raw result

        '''
        self.labels = pd.DataFrame(self.pipeline(text))

        self.labels['word'] = self.labels['word'].apply(lambda arr:list(arr.split(" ")))
        self.labels = self.labels.explode('word',ignore_index=True)
        columns = ['word','entity_group','score','start','end']

        if details == 'minimum':
            predicts = self.labels[columns[:2]]

        if details == "medium":
            predicts = self.labels[columns[:3]]

        if details == "all":
            predicts = self.labels[columns]

        # with pd.option_context('display.max_rows', 10,
        #                'display.max_columns', None,
        #                'display.width', 1000,
        #                'display.colheader_justify', 'left'):
        #     display(predicts)

        if as_dict:
            return predicts.to_dict('records')
        return predicts

    def getTokenLabels(self,text):
        predictions = self.pipeline(text)

        self.tokenLabels  = ""
        for token in predictions:
            subwords = list(token['word'].strip().split(" "))
            for _ in subwords:
                self.tokenLabels  = self.tokenLabels + (" "  + token['entity_group'])

        self.tokenLabels = self.tokenLabels.lstrip()
        return self.tokenLabels


    def listModels(self):
        print(" tagger models: ")
        for model in paths['tagger']:
            print("\t",model, ": ", paths['tagger'][model])
        for task in set(paths) - {'tagger'}:
            print("\n",task,"models: ")
            for model in paths[task]:
                print("\t",model, ": ", paths[task][model])



from transformers import AutoTokenizer, AutoModelForCausalLM
from transformers import pipeline
import pandas as pd
from ..config import paths

class GPTModel:
    def __init__(self, modelName='marathi-gpt'):
        self.modelName = modelName
        self.modelRoute = paths['autocomplete'][self.modelName]
        self.tokenizer = AutoTokenizer.from_pretrained(self.modelRoute)
        self.model = AutoModelForCausalLM.from_pretrained(self.modelRoute)
        self.classifier = pipeline('text-generation',
                              model=self.model, tokenizer=self.tokenizer)
        pd.options.display.max_colwidth = None 

    def nextWord(self, text, numOfPredictions = 1):
        result = self.classifier(text, max_new_tokens = 1, num_return_sequences = numOfPredictions)
        df = pd.DataFrame.from_dict(result)
        return df


    def completeSentence(self, text, numOfWords = 25, numOfPredictions = 1):
        result = self.classifier(text, max_new_tokens = numOfWords, num_return_sequences = numOfPredictions)
        df = pd.DataFrame.from_dict(result)
        return df

    def listModels(self):
        print(" autocomplete models: ")
        for model in paths['autocomplete']:
            print("\t",model, ": ", paths['autocomplete'][model])
        for task in set(paths) - {'autocomplete'}:
            print("\n",task,"models: ")
            for model in paths[task]:
                print("\t",model, ": ", paths[task][model])


from setuptools import setup

setup(
    name='randomlib',
    version='4.2',
    # author='L3Cube',
    # author_email='l3cube.pune@gmail.com',
    # description='An NLP Library for Marathi Language',
    # url='https://github.com/l3cube-pune/MarathiNLP.git',
    packages=['randomlib','randomlib.autocomplete', 'randomlib.datasets', 'randomlib.hate', 'randomlib.maskFill', 'randomlib.modelRepo', 'randomlib.preprocess', 'randomlib.sentiment', 'randomlib.similarity', 'randomlib.tagger', 'randomlib.tokenizer'],# can also use setuptools.find_packages()
    include_package_data=True,
    install_requires=['importlib_resources','huggingface_hub==0.11.1','tqdm','pandas','sentence_transformers','transformers','numpy','torch','IPython'], #'openpyxl'
    license='MIT',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from transformers import AutoTokenizer, AutoModelForMaskedLM, BertTokenizer, BertForMaskedLM
from transformers import pipeline
from IPython.display import display
import pandas as pd
from ..config import paths
#Sentence similarity
#sentence embedding
class MaskFillModel:
    def __init__(self, modelName='marathi-bert-v2'):
        self.modelName = modelName
        self.modelRoute = paths['maskFill'][self.modelName]
        self.task = "fill-mask" 

        if modelName == 'marathi-bert-v2':
            self.tokenizer = BertTokenizer.from_pretrained(self.modelRoute)
            self.model = BertForMaskedLM.from_pretrained(self.modelRoute)
        else:
            self.tokenizer = AutoTokenizer.from_pretrained(self.modelRoute)
            self.model = AutoModelForMaskedLM.from_pretrained(self.modelRoute)
        
        self.generator = pipeline(task = self.task,
                                    model=self.model,
                                    tokenizer=self.tokenizer)

    def predictMask(self,text: str,details: str = "minimum",as_dict: bool = False):
        if text.find(self.tokenizer.mask_token) == -1:
            print("Please mask your sentence first!")
            return None
        try:
            self.predictions = pd.DataFrame(self.generator(text))
        except Exception as e:
            print("Error in Model" + str(e))
        
        self.predictions['token_str'] = self.predictions['token_str'].apply(lambda word: word.replace(" ",""))

        if details == 'minimum':
            custom_predict = self.predictions[['token_str','sequence']]

        if details == "medium":
            custom_predict = self.predictions[['token_str','score','sequence']]

        if details == "all":
            custom_predict = self.predictions

        if as_dict:
            return custom_predict.to_dict('records')
        return custom_predict

    def listModels(self):
        print(" maskFill models: ")
        for model in paths['maskFill']:
            print("\t",model, ": ", paths['maskFill'][model])
        for task in set(paths) - {'maskFill'}:
            print("\n",task,"models: ")
            for model in paths[task]:
                print("\t",model, ": ", paths[task][model])


    



# from randomlib import setup


class Tokenize:
    def __init__(self,lang='mr'):
        self.lang = lang
        
##INTERNAL Function, not meant for programmer's usage
    def sentenceTokenize_mr(self, txt):
        punc_for_sentence_end = '''.!?'''
        sentences = []
        string = ""
        for i in txt:
            if i == "\n": continue
            if i not in punc_for_sentence_end:
                string += i
            else:
                string += i
                sentences.append(string)
                string = ""

        return sentences

    def sentenceTokenize(self, txt):
        if (self.lang == 'mr'):
            return self.sentenceTokenize_mr(txt)

##INTERNAL Function, not meant for programmer's usage
    def wordTokenize_mr(self, txt, punctuation):
        punc = '''!()-[]{};:'"\,<>./?@#$%^&*_~'''
        if punctuation:
            str = ""
            tokens = []
            for ele in txt:
                if ele in punc:
                    if str:
                        tokens.append(str)
                        str = ""
                    tokens.append(ele)
                elif ele == " ":
                    if str:
                        tokens.append(str)
                        str = ""
                else:
                    str += ele
            if str:
                tokens.append(str)
                str = ""
            return tokens
        else:
            for ele in txt:
                if ele in punc:
                    txt = txt.replace(ele, " ")
            x = txt.split()
            return x

    def wordTokenize(self, line, punctuation=True):
        if (self.lang == 'mr'):
            result = self.wordTokenize_mr(line, punctuation)
            return result
